from setuptools import setup, find_packages

setup(
    name="cagpaint",
    version="0.1.7",
    packages=find_packages(),
    install_requires=[
        # Add your package dependencies here, e.g.,
        # 'numpy>=1.20.0',
    ],
    author="ChristianEschen",
    author_email="christian_eschen@hotmail.com",
    description="cagpaint",
    url="https://github.com/yourusername/your-package"
)
